﻿using UnityEngine;
using System;
using System.Collections;
using System.Collections.Generic;

public class TextureMeshGenerator : MonoBehaviour
{
    public int newVertices;
    public bool colorVertices;

    public Material geometryMaterial;
    public Material textureMaterial;

    private int relaxationPhaseCount;
    private float radius;
    private float force;
    private float drag;
    private bool triangulated;

    private GameObject geometryContainer;
    private Mesh geometryMesh;

    private List<Vertex> textureVertices;
    private Dictionary<Vector3, Vertex> geometryVertices;
    private List<Triangle> geometryTriangles;

    private Dictionary<Vector3, float> vertexNeighbourhoodBuffer;

    ///<summary>
    ///Initialization.
    ///Expects a GameObject called "Geometry".
    ///</summary>
    void Start()
    {
        UnityEngine.Random.seed = 0;

        relaxationPhaseCount = 40;
        force = 0.5f;
        drag = 0.2f;
        triangulated = false;

        geometryContainer = GameObject.Find("Geometry");
        geometryMesh = geometryContainer.GetComponentInChildren<MeshFilter>().sharedMesh;

        textureVertices = new List<Vertex>();
        geometryVertices = new Dictionary<Vector3, Vertex>();
        geometryTriangles = new List<Triangle>();

        ProcessGeometryMesh();

        CreateTextureVertices();
    }

    ///<summary>
    ///Used for camera control in the Game view.
    ///</summary>
    void Update()
    {
        if(Input.GetButton("Fire1"))
        {
            float h = -2 * Input.GetAxis("Mouse X");
            float v = 2 * Input.GetAxis("Mouse Y");
            Vector3 rotVec = new Vector3(v, h, 0);
            rotVec = transform.InverseTransformDirection(rotVec);
            transform.Rotate(rotVec);
        }

        float z = -2 * Input.GetAxis("Mouse ScrollWheel");
        Vector3 traVec = new Vector3(0, 0, z);
        traVec = transform.InverseTransformDirection(traVec);
        transform.Translate(traVec);
    }

    ///<summary>
    ///Point repulsion happens here, as well as preparations for triangulation.
    ///</summary>
    void FixedUpdate()
    {
        ///<summary>
        ///40 iterations to make sure points are evenly spaced before trying to triangulate.
        ///Value arrived at through testing.
        ///</summary>
        while(relaxationPhaseCount > 0)
        {
            for (int i = 0; i < textureVertices.Count; i++)
            {
                textureVertices[i].vertexNeighbours = new List<Vertex>();

                textureVertices[i].radius = radius;
                textureVertices[i].force = force;
                textureVertices[i].drag = drag;

                List<Vertex> n = ExtractTextureVertices(textureVertices[i].parent, textureVertices[i].position);

                for (int j = 0; j < n.Count; j++)
                {
                    float distance = (textureVertices[i].position - n[j].position).magnitude;

                    if (distance <= radius && textureVertices[i].position != n[j].position)
                    {
                        textureVertices[i].vertexNeighbours.Add(n[j]);
                        textureVertices[i].AddForce(textureVertices[i].position - n[j].position);
                    }
                }

                textureVertices[i].Reposition();

                PointIndex closestPoint = ClosestPointOnGeometrySurface(textureVertices[i].parent, textureVertices[i].position);

                textureVertices[i].position = closestPoint.position;

                if (textureVertices[i].parent.listIndex != closestPoint.index)
                {
                    geometryTriangles[textureVertices[i].parent.listIndex].textureVertices.Remove(textureVertices[i]);
                    geometryTriangles[closestPoint.index].textureVertices.Add(textureVertices[i]);
                    textureVertices[i].parent = geometryTriangles[closestPoint.index];
                }
            }

            relaxationPhaseCount--;
        }

        ///<summary>
        ///Increases / decreases radii of vertices that are neighbour to holes / irregularities in the mesh.
        ///Once no more irregularities are found, the mesh is triangulated.
        ///</summary>
        if(!triangulated)
        {
            int crossCount = 0;
            int holeCount = 0;

            for (int i = 0; i < textureVertices.Count; i++)
            {
                textureVertices[i].vertexNeighbours = new List<Vertex>();

                textureVertices[i].force = force;
                textureVertices[i].drag = drag;

                for (int j = 0; j < textureVertices.Count; j++)
                {
                    float distance = (textureVertices[i].position - textureVertices[j].position).magnitude;

                    if (distance <= textureVertices[i].radius && i != j)
                    {
                        textureVertices[i].vertexNeighbours.Add(textureVertices[j]);
                        textureVertices[i].AddForce(textureVertices[i].position - textureVertices[j].position);
                    }
                }

                textureVertices[i].Reposition();

                PointIndex closestPoint = ClosestPointOnGeometrySurface(textureVertices[i].parent, textureVertices[i].position);

                textureVertices[i].position = closestPoint.position;

                if (textureVertices[i].parent.listIndex != closestPoint.index)
                {
                    geometryTriangles[textureVertices[i].parent.listIndex].textureVertices.Remove(textureVertices[i]);
                    geometryTriangles[closestPoint.index].textureVertices.Add(textureVertices[i]);
                    textureVertices[i].parent = geometryTriangles[closestPoint.index];
                }
            }

            for (int i = 0; i < textureVertices.Count; i++)
            {
                for (int j = 0; j < textureVertices[i].vertexNeighbours.Count; j++)
                {
                    Edge test = new Edge(textureVertices[i], textureVertices[i].vertexNeighbours[j]);

                    if (CleanUpEdge(test) == 0)
                    {
                        holeCount++;
                        textureVertices[i].radius += (textureVertices[i].radius * 0.0001f);
                    }
                    else if (CleanUpEdge(test) == 1)
                    {
                        crossCount++;
                        textureVertices[i].radius -= (textureVertices[i].radius * 0.0009f);
                    }
                }
            }

            if (holeCount == 0 && crossCount == 0 && !triangulated)
            {
                Triangulate();
            }
        }
    }

    ///<summary>
    ///Extracts mesh information from unity's data structures.
    ///</summary>
    private void ProcessGeometryMesh()
    {
        float surfaceArea = 0;

        for (int i = 0; i < geometryMesh.triangles.Length / 3; i++)
        {
            int vertexID1 = geometryMesh.triangles[i * 3];
            int vertexID2 = geometryMesh.triangles[i * 3 + 1];
            int vertexID3 = geometryMesh.triangles[i * 3 + 2];

            Vector3 position1 = geometryMesh.vertices[vertexID1];
            Vector3 position2 = geometryMesh.vertices[vertexID2];
            Vector3 position3 = geometryMesh.vertices[vertexID3];

            Vector3 normal = (geometryMesh.normals[vertexID1] + geometryMesh.normals[vertexID2] + geometryMesh.normals[vertexID3]) / 3;

            Vertex vertex1;
            Vertex vertex2;
            Vertex vertex3;

            if (!geometryVertices.ContainsKey(position1))
            {
                vertex1 = new Vertex(position1, vertexID1, radius, null);
                geometryVertices.Add(position1, vertex1);
            }
            else
            {
                geometryVertices[position1].indices.Add(vertexID1);
                vertex1 = geometryVertices[position1];
            }

            if (!geometryVertices.ContainsKey(position2))
            {
                vertex2 = new Vertex(position2, vertexID2, radius, null);
                geometryVertices.Add(position2, vertex2);
            }
            else
            {
                geometryVertices[position2].indices.Add(vertexID2);
                vertex2 = geometryVertices[position2];
            }

            if (!geometryVertices.ContainsKey(position3))
            {
                vertex3 = new Vertex(position3, vertexID3, radius, null);
                geometryVertices.Add(position3, vertex3);
            }
            else
            {
                geometryVertices[position3].indices.Add(vertexID3);
                vertex3 = geometryVertices[position3];
            }

            if (!vertex1.vertexNeighbours.Contains(vertex2))
            {
                vertex1.vertexNeighbours.Add(vertex2);
            }

            if (!vertex1.vertexNeighbours.Contains(vertex3))
            {
                vertex1.vertexNeighbours.Add(vertex3);
            }

            if (!vertex2.vertexNeighbours.Contains(vertex1))
            {
                vertex2.vertexNeighbours.Add(vertex1);
            }

            if (!vertex2.vertexNeighbours.Contains(vertex3))
            {
                vertex2.vertexNeighbours.Add(vertex3);
            }

            if (!vertex3.vertexNeighbours.Contains(vertex1))
            {
                vertex3.vertexNeighbours.Add(vertex1);
            }

            if (!vertex3.vertexNeighbours.Contains(vertex2))
            {
                vertex3.vertexNeighbours.Add(vertex2);
            }

            Triangle triangle = new Triangle(vertex1, vertex2, vertex3, normal, i * 3);

            vertex1.triangleNeighbours.Add(triangle);
            vertex2.triangleNeighbours.Add(triangle);
            vertex3.triangleNeighbours.Add(triangle);

            geometryTriangles.Add(triangle);

            surfaceArea += triangle.area;
        }

        radius = 1.25f * Mathf.Sqrt(surfaceArea / newVertices);

        geometryTriangles.Sort();

        float percentile = 0;

        for (int i = 0; i < geometryTriangles.Count; i++)
        {
            float percent = geometryTriangles[i].area / surfaceArea;

            percentile += percent;
            geometryTriangles[i].areaPercentile = percentile;
            geometryTriangles[i].listIndex = i;
        }

        geometryTriangles[geometryTriangles.Count - 1].areaPercentile = 1.0f;

        CalculateTriangleNeighbourhood(surfaceArea);

        geometryContainer.GetComponentInChildren<MeshRenderer>().material = geometryMaterial;
    }

    ///<summary>
    ///Determines every neighbouring triangle within a certain distance for every entry of "geometryTriangles".
    ///</summary>
    ///<param name="surfaceArea">
    ///Surface area of the input mesh.
    ///Used to calculate the distance mentioned in the summary.
    ///</param>
    private void CalculateTriangleNeighbourhood(float surfaceArea)
    {
        float neighbourhoodRadius = 2 * (surfaceArea / newVertices);

        Dictionary<int, Triangle> triangleNeighbourhoodBuffer;

        for (int i = 0; i < geometryTriangles.Count; i++)
        {
            vertexNeighbourhoodBuffer = new Dictionary<Vector3, float>();

            vertexNeighbourhoodBuffer.Add(geometryTriangles[i].a.position, 0.0f);
            vertexNeighbourhoodBuffer.Add(geometryTriangles[i].b.position, 0.0f);
            vertexNeighbourhoodBuffer.Add(geometryTriangles[i].c.position, 0.0f);

            triangleNeighbourhoodBuffer = new Dictionary<int, Triangle>();

            triangleNeighbourhoodBuffer.Add(geometryTriangles[i].index, geometryTriangles[i]);

            CalculateVertexNeighbourhood(geometryTriangles[i].a, geometryTriangles[i].a, neighbourhoodRadius);
            CalculateVertexNeighbourhood(geometryTriangles[i].b, geometryTriangles[i].b, neighbourhoodRadius);
            CalculateVertexNeighbourhood(geometryTriangles[i].c, geometryTriangles[i].c, neighbourhoodRadius);

            foreach (Vector3 v in vertexNeighbourhoodBuffer.Keys)
            {
                geometryTriangles[i].geometryVertexNeighbourhood.Add(geometryVertices[v]);

                foreach(Triangle t in geometryVertices[v].triangleNeighbours)
                {
                    if(!triangleNeighbourhoodBuffer.ContainsKey(t.index))
                    {
                        triangleNeighbourhoodBuffer.Add(t.index, t);
                    }
                }
            }

            foreach(int j in triangleNeighbourhoodBuffer.Keys)
            {
                geometryTriangles[i].triangleNeighbourhood.Add(triangleNeighbourhoodBuffer[j]);
            }
        }
    }

    ///<summary>
    ///Searches for vertices within a range of a source vertex by stepping through it's adjacent points.
    ///</summary>
    ///<param name="source">
    ///The neighbourhood of this vertex is calculated.
    ///</param>
    ///<param name="neighbour">
    ///Used for recursion.
    ///Initially set to source as well.
    ///</param>
    ///<param name="neighbourhoodRadius">
    ///The maximum Euclidean distance a neighbouring vertex is allowed to be away from source.
    ///</param>
    private void CalculateVertexNeighbourhood(Vertex neighbour, Vertex source, float neighbourhoodRadius)
    {
        for (int i = 0; i < neighbour.vertexNeighbours.Count; i++)
        {
            float distanceToSource = (source.position - neighbour.vertexNeighbours[i].position).magnitude;

            if (distanceToSource <= neighbourhoodRadius)
            {
                if (!vertexNeighbourhoodBuffer.ContainsKey(neighbour.vertexNeighbours[i].position))
                {
                    vertexNeighbourhoodBuffer.Add(neighbour.vertexNeighbours[i].position, distanceToSource);
                    CalculateVertexNeighbourhood(neighbour.vertexNeighbours[i], source, neighbourhoodRadius);
                }
            }
        }
    }

    ///<summary>
    ///Randomly generates texture vertices within triangles.
    ///Triangles with bigger areas have a higher chance to contain a new vertex.
    ///</summary>
    private void CreateTextureVertices()
    {
        for (int i = 0; i < newVertices; i++)
        {
            float randomTriangle = UnityEngine.Random.value;

            Triangle target = geometryTriangles.Find(x => x.areaPercentile >= randomTriangle);

            Vector3 randomPosition = GetRandomPointInTriangle(target);

            Vertex newVertex = new Vertex(randomPosition, i, radius, target);

            textureVertices.Add(newVertex);

            target.textureVertices.Add(newVertex);
        }
    }

    ///<summary>
    ///Returns a random point inside the triangle.
    ///</summary>
    private Vector3 GetRandomPointInTriangle(Triangle target)
    {
        float s = UnityEngine.Random.value;
        float t = UnityEngine.Random.value;

        float x = 1 - Mathf.Sqrt(t);
        float y = (1 - s) * Mathf.Sqrt(t);
        float z = s * Mathf.Sqrt(t);

        Vector3 randomPoint = target.a.position * x + target.b.position * y + target.c.position * z;

        return randomPoint;
    }

    ///<summary>
    ///Returns every texture vertex the triangle and it's neighbourhood contain.
    ///</summary>
    private List<Vertex> ExtractTextureVertices(Triangle triangle, Vector3 notIncluded)
    {
        List<Vertex> textureVertices = new List<Vertex>();

        foreach(Vertex v in triangle.textureVertices)
        {
            if(v.position != notIncluded)
            {
                textureVertices.Add(v);
            }
        }

        foreach(Triangle t in triangle.triangleNeighbourhood)
        {
            foreach(Vertex v in t.textureVertices)
            {
                if (v.position != notIncluded)
                {
                    textureVertices.Add(v);
                }
            }
        }

        return textureVertices;
    }

    ///<summary>
    ///Returns closest point on a mesh if measured from a given point P.
    ///Checks every triangle in a neighbourhood.
    ///</summary>
    ///<param name="parent">
    ///The triangle last containing P.
    ///</param>
    ///<param name="position">
    ///The position of P.
    ///</param>
    ///<returns>
    ///The closest point and the index of the triangle containing said point.
    ///</returns>
    private PointIndex ClosestPointOnGeometrySurface(Triangle parent, Vector3 position)
    {
        Vertex closest = parent.geometryVertexNeighbourhood[0];

        for (int i = 1; i < parent.geometryVertexNeighbourhood.Count; i++)
        {
            if ((closest.position - position).magnitude > (parent.geometryVertexNeighbourhood[i].position - position).magnitude)
            {
                closest = parent.geometryVertexNeighbourhood[i];
            }
        }

        Vector3 closestPointOnSurface = ClosestPointOnTriangle(closest.triangleNeighbours[0], position);
        int closestTriangleIndex = closest.triangleNeighbours[0].listIndex;

        for (int i = 1; i < closest.triangleNeighbours.Count; i++)
        {
            if ((position - closestPointOnSurface).magnitude > (position - ClosestPointOnTriangle(closest.triangleNeighbours[i], position)).magnitude)
            {
                closestPointOnSurface = ClosestPointOnTriangle(closest.triangleNeighbours[i], position);
                closestTriangleIndex = closest.triangleNeighbours[i].listIndex;
            }
        }

        return new PointIndex(closestPointOnSurface, closestTriangleIndex);
    }

    ///<summary>
    ///Returns closest point on a triangle if measured from a given point P.
    ///</summary>
    private Vector3 ClosestPointOnTriangle(Triangle triangle, Vector3 P)
    {
        Vector3 B, E0, E1, D;
        float a, b, c, d, e, delta, denominator;

        B = triangle.a.position;
        E0 = triangle.b.position - B;
        E1 = triangle.c.position - B;

        a = Vector3.Dot(E0, E0);
        b = Vector3.Dot(E0, E1);
        c = Vector3.Dot(E1, E1);

        delta = a * c - b * b;
        denominator = a - 2 * b + c;

        D = B - P;
        d = Vector3.Dot(E0, D);
        e = Vector3.Dot(E1, D);

        float s = b * e - c * d;
        float t = b * d - a * e;

        if (s + t <= delta)
        {
            if (s < 0)
            {
                if (t < 0)
                {
                    float tmp0 = d;
                    float tmp1 = e;

                    if (tmp1 > tmp0)
                    {
                        s = tmp0 <= 0 ? 1 : d >= 0 ? 0 : -d / a;
                        t = 0;
                    }
                    else
                    {
                        s = 0;
                        t = tmp1 <= 0 ? 1 : e >= 0 ? 0 : -e / c;
                    }
                }
                else
                {
                    s = 0;
                    t = e >= 0 ? 0 : -e >= c ? 1 : -e / c;
                }
            }
            else if (t < 0)
            {
                s = d >= 0 ? 0 : -d >= a ? 1 : -d / a;
                t = 0;
            }
            else
            {
                float inverseDelta = 1 / delta;
                s *= inverseDelta;
                t *= inverseDelta;
            }
        }
        else
        {
            if (s < 0)
            {
                float tmp0 = b + d;
                float tmp1 = c + e;

                if (tmp1 > tmp0)
                {
                    float numerator = tmp1 - tmp0;
                    s = numerator >= denominator ? 1 : numerator / denominator;
                    t = 1 - s;
                }
                else
                {
                    s = 0;
                    t = tmp1 <= 0 ? 1 : e >= 0 ? 0 : -e / c;
                }
            }
            else if (t < 0)
            {
                float tmp0 = b + e;
                float tmp1 = a + d;

                if (tmp1 > tmp0)
                {
                    float numerator = tmp1 - tmp0;
                    s = numerator >= denominator ? 1 : numerator / denominator;
                    t = 1 - s;
                }
                else
                {
                    s = tmp1 <= 0 ? 1 : d >= 0 ? 0 : -d / a;
                    t = 0;
                }
            }
            else
            {
                float numerator = c + e - b - d;

                if (numerator <= 0)
                {
                    s = 0;
                }
                else
                {
                    s = numerator >= denominator ? 1 : numerator / denominator;
                }

                t = 1 - s;
            }
        }

        Vector3 Q = B + s * E0 + t * E1;

        return Q;
    }

    ///<summary>
    ///Triangulates the texture vertices.
    ///Applies normal, uv / color.
    ///</summary>
    private void Triangulate()
    {
        triangulated = true;

        GameObject textureContainer = Instantiate(geometryContainer, transform.position, transform.rotation) as GameObject;
        textureContainer.transform.parent = this.transform;
        textureContainer.name = "Texture";
        textureContainer.GetComponentInChildren<MeshRenderer>().material = textureMaterial;

        List<Triangle> textureTriangles = new List<Triangle>();

        if(colorVertices)
        {
            Vector3[] textureMeshVertices = new Vector3[textureVertices.Count];
            Vector3[] textureMeshNormals = new Vector3[textureVertices.Count];
            Color[] textureMeshColors = new Color[textureVertices.Count];
        
            for(int i = 0; i < textureVertices.Count; i++)
            {
                textureVertices[i].indices[0] = i;

                textureMeshVertices[i] = textureVertices[i].position;
                textureMeshNormals[i] = textureVertices[i].parent.normal;

                float r = UnityEngine.Random.value;
                Color col = Color.white;

                if (r < 0.33)
                {
                    col = Color.red;
                }
                else if (r < 0.66)
                {
                    col = Color.green;
                }
                else
                {
                    col = Color.blue;
                }

                textureMeshColors[i] = col;

                Vertex a = textureVertices[i];
                Vertex b;
                Vertex c;

                for(int j = 0; j < a.vertexNeighbours.Count; j++)
                {
                    b = a.vertexNeighbours[j];

                    for(int k = 0; k < a.vertexNeighbours.Count; k++)
                    {
                        if(k != j)
                        {
                            for(int l = 0; l < b.vertexNeighbours.Count; l++)
                            {
                                if(a.vertexNeighbours[k].position == b.vertexNeighbours[l].position)
                                {
                                    c = a.vertexNeighbours[k];
                                    Triangle triangle = new Triangle(a, b, c, a.parent.normal, -1);
                                    textureTriangles.Add(triangle);
                                    a.triangleNeighbours.Add(triangle);
                                    b.triangleNeighbours.Add(triangle);
                                    c.triangleNeighbours.Add(triangle);
                                    b.vertexNeighbours.Remove(a);
                                    c.vertexNeighbours.Remove(a);
                                }
                            }
                        }
                    }
                }
            }

            int[] textureMeshTriangles = new int[textureTriangles.Count * 3];

            for(int i = 0; i < textureTriangles.Count; i++)
            {
                textureMeshTriangles[i * 3] = textureTriangles[i].a.indices[0];
                textureMeshTriangles[i * 3 + 1] = textureTriangles[i].b.indices[0];
                textureMeshTriangles[i * 3 + 2] = textureTriangles[i].c.indices[0];
            }

            Mesh textureMesh = new Mesh();
            textureMesh.vertices = textureMeshVertices;
            textureMesh.normals = textureMeshNormals;
            textureMesh.triangles = textureMeshTriangles;
            textureMesh.colors = textureMeshColors;

            textureContainer.GetComponentInChildren<MeshFilter>().mesh = textureMesh;
        }
        else
        {
            for (int i = 0; i < textureVertices.Count; i++)
            {
                textureVertices[i].indices[0] = i;

                Vertex a = textureVertices[i];
                Vertex b;
                Vertex c;

                for (int j = 0; j < a.vertexNeighbours.Count; j++)
                {
                    b = a.vertexNeighbours[j];

                    for (int k = 0; k < a.vertexNeighbours.Count; k++)
                    {
                        if (k != j)
                        {
                            for (int l = 0; l < b.vertexNeighbours.Count; l++)
                            {
                                if (a.vertexNeighbours[k].position == b.vertexNeighbours[l].position)
                                {
                                    c = a.vertexNeighbours[k];
                                    Triangle triangle = new Triangle(a, b, c, a.parent.normal, -1);
                                    textureTriangles.Add(triangle);
                                    a.triangleNeighbours.Add(triangle);
                                    b.triangleNeighbours.Add(triangle);
                                    c.triangleNeighbours.Add(triangle);
                                    b.vertexNeighbours.Remove(a);
                                    c.vertexNeighbours.Remove(a);
                                }
                            }
                        }
                    }
                }
            }

            int[] textureMeshTriangles = new int[textureTriangles.Count * 3];
            Vector3[] textureMeshVertices = new Vector3[textureTriangles.Count * 3];
            Vector3[] textureMeshNormals = new Vector3[textureTriangles.Count * 3];
            Vector2[] textureMeshUVs = new Vector2[textureTriangles.Count * 3];
            Color[] textureMeshColors = new Color[textureTriangles.Count * 3];

            for (int i = 0; i < textureTriangles.Count; i++)
            {
                textureMeshVertices[i * 3] = textureTriangles[i].a.position;
                textureMeshVertices[i * 3 + 1] = textureTriangles[i].b.position;
                textureMeshVertices[i * 3 + 2] = textureTriangles[i].c.position;

                Vector3 newNormal = (textureTriangles[i].a.parent.normal + textureTriangles[i].a.parent.normal + textureTriangles[i].a.parent.normal) / 3;

                textureMeshNormals[i * 3] = newNormal;
                textureMeshNormals[i * 3 + 1] = newNormal;
                textureMeshNormals[i * 3 + 2] = newNormal;

                float r = UnityEngine.Random.value;
                float y = Mathf.Sqrt(3) / 2;

                if (r < 0.25)
                {
                    textureMeshUVs[i * 3] = new Vector2(0, 0);
                    textureMeshUVs[i * 3 + 1] = new Vector2(0.25f, 0);
                    textureMeshUVs[i * 3 + 2] = new Vector2(0.125f, y);
                }
                else if (r < 0.5)
                {
                    textureMeshUVs[i * 3] = new Vector2(0.25f, 0);
                    textureMeshUVs[i * 3 + 1] = new Vector2(0.5f, 0);
                    textureMeshUVs[i * 3 + 2] = new Vector2(0.375f, y);
                }
                else if (r < 0.75)
                {
                    textureMeshUVs[i * 3] = new Vector2(0.5f, 0);
                    textureMeshUVs[i * 3 + 1] = new Vector2(0.75f, 0);
                    textureMeshUVs[i * 3 + 2] = new Vector2(0.625f, y);
                }
                else
                {
                    textureMeshUVs[i * 3] = new Vector2(0.75f, 0);
                    textureMeshUVs[i * 3 + 1] = new Vector2(1, 0);
                    textureMeshUVs[i * 3 + 2] = new Vector2(0.875f, y);
                }

                textureMeshColors[i * 3] = Color.white;
                textureMeshColors[i * 3 + 1] = Color.white;
                textureMeshColors[i * 3 + 2] = Color.white;

                textureMeshTriangles[i * 3] = i * 3;
                textureMeshTriangles[i * 3 + 1] = i * 3 + 1;
                textureMeshTriangles[i * 3 + 2] = i * 3 + 2;
            }

            Mesh textureMesh = new Mesh();
            textureMesh.vertices = textureMeshVertices;
            textureMesh.normals = textureMeshNormals;
            textureMesh.uv = textureMeshUVs;
            textureMesh.triangles = textureMeshTriangles;
            textureMesh.colors = textureMeshColors;

            textureContainer.GetComponentInChildren<MeshFilter>().mesh = textureMesh;
        }
    }

    ///<summary>
    ///Used to inspect edges for connected holes.
    ///</summary>
    ///<returns>
    ///0 if the edge is connected to a square hole.
    ///1 if the edge is connected to a filled square hole (intersecting edges).
    ///2 if the edge is valid.
    ///</returns>
    private int CleanUpEdge(Edge edge)
    {
        int neighbours = CalculateEdgeNeighbours(edge);

        if(neighbours < 2)
        {
            Debug.DrawLine(edge.a.position, edge.b.position, Color.green);
            return 0;
        }
        else if(neighbours > 2)
        {
            Debug.DrawLine(edge.a.position, edge.b.position, Color.red);
            return 1;
        }
        else
        {
            Debug.DrawLine(edge.a.position, edge.b.position, Color.blue);
            return 2;
        }
    }

    ///<summary>
    ///Calculates the number of vertices that contain both of an edge's endpoints in their neighbourhood list.
    ///</summary>
    ///<returns>
    ///Number of directly connected vertices.
    ///</returns>
    private int CalculateEdgeNeighbours(Edge edge)
    {
        int neighbours = 0;

        for(int i = 0; i < edge.a.vertexNeighbours.Count; i++)
        {
            if(edge.a.position != edge.a.vertexNeighbours[i].position)
            {
                for (int j = 0; j < edge.b.vertexNeighbours.Count; j++)
                {
                    if (edge.b.position != edge.b.vertexNeighbours[j].position && edge.a.vertexNeighbours[i].position == edge.b.vertexNeighbours[j].position)
                    {
                        neighbours++;
                    }
                }
            }
        }

        return neighbours;
    }
}

///<summary>
///Class representing a vertex.
///</summary>
public class Vertex : IComparable<Vertex>
{
    public Vector3 position;
    public float radius;
    public Triangle parent;

    public float force;
    public float drag;

    public List<Vertex> vertexNeighbours;
    public List<Triangle> triangleNeighbours;
    public List<int> indices;

    public float neighbourDistance;

    private Vector3 forceDirection;

    public Vertex(Vector3 position, int index, float radius, Triangle parent)
    {
        this.position = position;
        this.radius = radius;
        this.parent = parent;

        force = 0.1f;
        drag = 0.0f;

        vertexNeighbours = new List<Vertex>();
        triangleNeighbours = new List<Triangle>();
        indices = new List<int>();
        indices.Add(index);

        neighbourDistance = 0.0f;

        forceDirection = Vector3.zero;
    }

    public int CompareTo(Vertex other)
    {
        if (other == null)
        {
            return 1;
        }

        if (other.neighbourDistance < neighbourDistance)
        {
            return 1;
        }
        else if (other.neighbourDistance > neighbourDistance)
        {
            return -1;
        }
        else
        {
            return 0;
        }
    }

    ///<summary>
    ///A force in the form of a direction in 3D space is added to the accumulated force-vector.
    ///</summary>
    ///<param name = "direction">
    ///The direction from which the force is applied. Magnitude indirectly proportional to strength of force.
    ///</param>
    public void AddForce(Vector3 direction)
    {
        forceDirection += (radius - direction.magnitude) * force * direction.normalized;
    }

    ///<summary>
    ///Changes position by adding accumulated force-vector to current position.
    ///Drag is simulated to reduce force gradually over time instead of setting it to zero after each call.
    ///</summary>
    public void Reposition()
    {
        position += forceDirection;
        forceDirection *= drag;
    }
}

///<summary>
///Container class for two neighbouring vertices.
///</summary>
public class Edge
{
    public Vertex a, b;

    public Edge(Vertex a, Vertex b)
    {
        this.a = a;
        this.b = b;
    }
}

///<summary>
///Class representing a triangle.
///</summary>
public class Triangle : IComparable<Triangle>
{
    public Vertex a, b, c;
    public Vector3 normal;
    public int index;
    public int listIndex;
    public float area;
    public float areaPercentile;

    public List<Vertex> geometryVertexNeighbourhood;
    public List<Vertex> textureVertices;
    public List<Triangle> triangleNeighbourhood;

    public Triangle(Vertex a, Vertex b, Vertex c, Vector3 normal, int index)
    {
        this.a = a;
        this.b = b;
        this.c = c;
        this.normal = normal;
        this.index = index;
        this.listIndex = -1;
        this.area = CalculateArea(a, b, c);
        this.areaPercentile = 0.0f;

        geometryVertexNeighbourhood = new List<Vertex>();
        geometryVertexNeighbourhood.Add(a);
        geometryVertexNeighbourhood.Add(b);
        geometryVertexNeighbourhood.Add(c);

        textureVertices = new List<Vertex>();
        triangleNeighbourhood = new List<Triangle>();
    }

    public int CompareTo(Triangle other)
    {
        if (other == null)
        {
            return 1;
        }

        if (other.area < area)
        {
            return -1;
        }
        else if (other.area > area)
        {
            return 1;
        }
        else
        {
            return 0;
        }
    }

    ///<summary>
    ///Method to find the edge opposite to a vertex.
    ///</summary>
    ///<param name = "vertex">
    ///Must be part of the triangle.
    ///</param>
    ///<returns>
    ///An edge containing the two opposite vertices.
    ///</returns>
    public Edge OppositeEdge(Vertex vertex)
    {
        if(vertex.position == a.position)
        {
            return new Edge(b, c);
        }
        else if(vertex.position == b.position)
        {
            return new Edge(a, c);
        }
        else if(vertex.position == c.position)
        {
            return new Edge(a, b);
        }

        return null;
    }

    private float CalculateArea(Vertex pointA, Vertex pointB, Vertex pointC)
    {
        Vector3 side1 = pointB.position - pointA.position;
        Vector3 side2 = pointC.position - pointA.position;

        return Mathf.Abs(Vector3.Cross(side1, side2).magnitude) * 0.5f;
    }
}

///<summary>
///Helper class. Used to store a position and an index together.
///</summary>
public struct PointIndex
{
    public Vector3 position;
    public int index;

    public PointIndex(Vector3 position, int index)
    {
        this.position = position;
        this.index = index;
    }
}
